#!/usr/bin/env python3
# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""The Composer CLI launcher for distributed training."""

import contextlib
import datetime
import logging
import os
import signal
import subprocess
import sys
import tempfile
import time
import traceback
from argparse import ArgumentParser
from typing import Any, Union

import psutil
import torch
from packaging import version

import composer
from composer.loggers.mosaicml_logger import (
    MOSAICML_GPU_LOG_FILE_PREFIX_ENV_VAR,
    MOSAICML_LOG_DIR_ENV_VAR,
    MOSAICML_PLATFORM_ENV_VAR,
)
from composer.utils import get_free_tcp_port

CLEANUP_TIMEOUT = datetime.timedelta(seconds=30)

log = logging.getLogger(__name__)


def _get_parser():
    parser = ArgumentParser(description='Utility for launching distributed machine learning jobs.')

    parser.add_argument('--version', action='version', version=f'MosaicML Composer {composer.__version__}')

    required_args = parser.add_argument_group('required arguments')

    parser.add_argument(
        '-n',
        '--nproc',
        type=int,
        help=(
            'The number of processes to launch on this node. Overrides env var `LOCAL_WORLD_SIZE` if specified; '
            'otherwise, defaults to `max(1, torch.cuda.device_count())`.'
        ),
    )

    parser.add_argument(
        '--stdout',
        type=str,
        default=None,
        help=(
            'Format string for a filename to dump the STDOUT from the non-local-rank-zero processes. '
            'The local rank zero process will be piped through to STDOUT. The available format variables are: '
            "'{rank}', '{local_rank}', '{world_size}', '{node_rank}', and '{local_world_size}'. If specified, "
            "it is recommended to include '{rank}' or '{local_rank}' in the filename so each rank will write to its "
            'own file. By default, the STDOUT of the non-local-rank-zero processes is discarded; instead, use the '
            'FileLogger within Composer. This logger captures and saves the STDOUT of each process.'
        ),
    )
    parser.add_argument(
        '--stderr',
        type=str,
        default=None,
        help=(
            'Format string for a filename to dump the STDERR from the non-local-rank-zero processes. '
            'The local rank zero process will be piped through to STDERR. The available format variables are: '
            "'{rank}', '{local_rank}', '{world_size}', '{node_rank}', and '{local_world_size}'. If specified, "
            "it is recommended to include '{rank}' or '{local_rank}' in the filename so each rank will write to its "
            'own file. By default, the STDERR of the non-local-rank-zero processes is discarded; instead, use the '
            'FileLogger within Composer. This logger captures and saves the STDERR of each process.'
        ),
    )
    parser.add_argument('-v', '--verbose', action='store_true', help='If set, print verbose messages')
    parser.add_argument(
        '-m',
        '--module_mode',
        action='store_true',
        help=(
            'If set, run the training script as a module instead of as a script. '
            'Cannot be used in conjunction with `command_mode`'
        ),
    )
    parser.add_argument(
        '-c',
        '--command_mode',
        action='store_true',
        help=(
            'If set, run the training script as a command (i.e. without `python`). '
            'Cannot be used in conjunction with `module_mode`.'
        ),
    )

    multinode_args = parser.add_argument_group(
        'multi-node arguments',
        description=(
            'These arguments generally only need to be set when training in a multi-node '
            'environment, i.e. when the world_size is bigger than nproc.'
        ),
    )
    multinode_args.add_argument(
        '--world_size',
        type=int,
        help=(
            'The total number of processes to launch across all nodes. '
            'Setting this to a value greater than nproc indicates a multi-node '
            'environment. Overrides env var WORLD_SIZE. Defaults to nproc.'
        ),
    )
    multinode_args.add_argument(
        '--base_rank',
        type=int,
        help=(
            'The rank of the lowest ranked process to launch on this node. '
            'Specifying a base_rank B and an nproc N will spawn processes with '
            'global ranks [B, B+1, ... B+N-1]. In a multi-node environment, '
            'at least one of base_rank and node_rank must be specified. '
            'If only one of base_rank and node_rank are provided, it is assumed '
            'that all nodes have the same amount of processes, and that the two '
            'values are related as node_rank * nproc = base_rank. If this is '
            'not the case, both base_rank and node_rank must be provided. '
            'Overrides env var BASE_RANK. Defaults to 0 in a single-node '
            'environment.'
        ),
    )
    multinode_args.add_argument(
        '--node_rank',
        type=int,
        help=(
            'The rank of this node. See base_rank for information on when '
            'this must be provided. Overrides env var NODE_RANK. Defaults to 0 '
            'in a single-node environment.'
        ),
    )
    multinode_args.add_argument(
        '--master_addr',
        type=str,
        help=(
            'The FQDN of the node hosting the C10d TCP store. For single-node '
            'operation, this can generally be left as 127.0.0.1. Overrides env var '
            'MASTER_ADDR. Defaults to 127.0.0.1 in a single-node environment.'
        ),
    )
    multinode_args.add_argument(
        '--master_port',
        type=int,
        help=(
            'The port on the master hosting the C10d TCP store. If you are '
            'running multiple trainers on a single node, this generally needs '
            'to be unique for each one. Overrides env var MASTER_PORT. Defaults '
            'to a random free port in a single-node environment.'
        ),
    )

    required_args.add_argument(
        'training_script',
        type=str,
        help=(
            'The path to the training script used to initialize a single training '
            'process. Should be followed by any command-line arguments the script '
            'should be launched with.'
        ),
    )
    required_args.add_argument(
        'training_script_args',
        nargs='...',
        help='Any arguments for the training script, given in the expected order.',
    )

    return parser


def _parse_args():
    parser = _get_parser()

    args = parser.parse_args()

    # Default values to env vars if they are not provided
    if args.nproc is None:
        if 'LOCAL_WORLD_SIZE' in os.environ:
            args.nproc = int(os.environ['LOCAL_WORLD_SIZE'])
        else:
            args.nproc = torch.cuda.device_count()

        if args.nproc == 0:
            # This could happen if doing cpu-only training,
            # which could cause torch.cuda.device_count() to return 0,
            # and LOCAL_WORLD_SIZE (as set by MCLI) to be zero
            args.nproc = 1

    if args.nproc < 1:
        raise ValueError('The nproc must be 1 or greater')

    if args.world_size is None:
        if 'WORLD_SIZE' in os.environ and os.environ.get('LOCAL_WORLD_SIZE') != os.environ['WORLD_SIZE']:
            # Use WORLD_SIZE env var if set and running multinode. Otherwise, default to nproc
            # to enable easy overriding of number of processes when on a single node.
            args.world_size = int(os.environ['WORLD_SIZE'])
        else:
            args.world_size = args.nproc

    if args.base_rank is None and 'BASE_RANK' in os.environ:
        args.base_rank = int(os.environ['BASE_RANK'])

    if args.node_rank is None and 'NODE_RANK' in os.environ:
        args.node_rank = int(os.environ['NODE_RANK'])

    if args.master_addr is None and 'MASTER_ADDR' in os.environ:
        args.master_addr = os.environ['MASTER_ADDR']

    if args.master_port is None and 'MASTER_PORT' in os.environ:
        args.master_port = int(os.environ['MASTER_PORT'])

    if args.world_size < args.nproc:
        raise ValueError(f'world_size({args.world_size}) cannot be less than nproc({args.nproc})')

    if args.world_size < 1:
        raise ValueError('The world_size must be 1 or greater')

    is_multinode = args.world_size > args.nproc

    if is_multinode:
        if args.base_rank is None and args.node_rank is None:
            raise ValueError(f'In a multi-node environment, at least one of node_rank and base_rank must be provided.')

        if args.node_rank is None:
            if args.world_size % args.nproc != 0 or args.base_rank % args.nproc != 0:
                raise ValueError(
                    'node_rank not provided, but unable to infer from base_rank since nodes appear to '
                    'have different amounts of processes. Please also specify node_rank.',
                )
            args.node_rank = args.base_rank // args.nproc

        if args.base_rank is None:
            if args.world_size % args.nproc != 0:
                raise ValueError(
                    'base_rank not provided, but unable to infer from node_rank since nodes appear to '
                    'have different amounts of processes. Please also provide base_rank.',
                )
            args.base_rank = args.node_rank * args.nproc

        if args.base_rank + args.nproc > args.world_size:
            raise ValueError(
                f'Cannot initialize processes for node with base_rank({args.base_rank}) and '
                f'nproc({args.nproc}) because this would mean creating a process with '
                f'rank({args.base_rank + args.nproc - 1}), and all processes must have smaller rank than '
                f'the world_size({args.world_size}).',
            )

        if args.master_addr is None:
            raise ValueError('In a multi-node environment, master_addr is required.')

        if args.master_port is None:
            raise ValueError('In a multi-node environment, master_port is required.')

    else:
        if args.base_rank is not None and args.base_rank != 0:
            raise ValueError(f'base_rank({args.base_rank}) != 0 is not valid in a single-node environment.')
        args.base_rank = 0

        if args.node_rank is not None and args.node_rank != 0:
            raise ValueError(f'node_rank({args.node_rank}) != 0 is not valid in a single-node environment.')
        args.node_rank = 0

        if args.master_addr is None:
            args.master_addr = '127.0.0.1'

        if args.master_port is None:
            args.master_port = get_free_tcp_port()

    return args


@contextlib.contextmanager
def _patch_env(**environs: str):
    """Returns a context manager that patches ``os.environ`` with ``environs``.

    The original ``os.environ`` values are restored at the end.
    """
    # Adapted loosely from https://stackoverflow.com/a/34333710
    # Capture the original environ values
    original_environs = {k: os.environ.get(k) for k in environs}

    # Patch the environment
    for k, v in environs.items():
        os.environ[k] = v
    try:
        # Run the context manager
        yield
    finally:
        # Restore the original environ values
        for k, v in original_environs.items():
            if v is None:
                del os.environ[k]
            else:
                os.environ[k] = v


def _launch_processes(
    nproc: int,
    world_size: int,
    base_rank: int,
    node_rank: int,
    master_addr: str,
    master_port: int,
    module_mode: bool,
    command_mode: bool,
    training_script: str,
    stdout_file_format: str,
    stderr_file_format: Union[str, None],
    training_script_args: list[Any],
    processes: dict[int, subprocess.Popen],
):
    log.info('Starting distributed environment on local node for global_rank(%s-%s)', base_rank, base_rank + nproc - 1)
    log.info('Distributed KV store: tcp://%s:%s', master_addr, master_port)

    nccl_env_variable = {
        (
            'NCCL_ASYNC_ERROR_HANDLING' if version.parse(torch.__version__) < version.parse('2.2.0') else 'TORCH_NCCL_ASYNC_ERROR_HANDLING'
        ):
            '1',
    }

    for local_rank in range(nproc):
        global_rank = base_rank + local_rank
        if command_mode and module_mode:
            raise ValueError('Either `command_mode` or `module_mode` should be set, but not both.')
        cmd = []
        if not command_mode:
            cmd.append(sys.executable)
        if module_mode:
            cmd.append('-m')

        cmd.append(training_script)

        # Update the env with the distributed variables
        with _patch_env(
            RANK=str(global_rank),
            WORLD_SIZE=str(world_size),
            LOCAL_RANK=str(local_rank),
            LOCAL_WORLD_SIZE=str(nproc),
            NODE_RANK=str(node_rank),
            MASTER_ADDR=master_addr,
            MASTER_PORT=str(master_port),
            PYTHONUNBUFFERED='1',
            **nccl_env_variable,
        ):
            # Populate the distributed variables in all launcher args
            for arg in training_script_args:
                cmd.append(os.path.expandvars(os.path.expanduser(arg)))

            log.info(
                'Launching process for local_rank(%s), global_rank(%s) with command(%s)',
                local_rank,
                global_rank,
                cmd,
            )

            if local_rank == 0:
                process = subprocess.Popen(
                    cmd,
                    text=True,
                )
            else:

                def _get_file(format: str):
                    filename = format.format(
                        rank=global_rank,
                        world_size=world_size,
                        local_rank=local_rank,
                        local_world_size=nproc,
                        node_rank=node_rank,
                    )
                    return open(filename, 'a+')

                stdout_file = _get_file(stdout_file_format)
                stderr_file = _get_file(stderr_file_format) if stderr_file_format is not None else None

                process = subprocess.Popen(
                    cmd,
                    stdout=stdout_file,
                    stderr=stderr_file if stderr_file is not None else subprocess.STDOUT,
                    text=True,
                )
                process.stdout = stdout_file
                if stderr_file is not None:
                    process.stderr = stderr_file
            processes[global_rank] = process


def _monitor_processes(processes: dict[int, subprocess.Popen]):
    try:
        while True:
            process_has_crashed = False
            all_processes_finished = True
            for global_rank, process in processes.items():
                if process.poll() is None:
                    # the process is still running
                    all_processes_finished = False
                    continue
                else:
                    # return code of 0 implies clean exit
                    if process.returncode != 0:
                        log.error(f'Rank {global_rank} crashed with exit code {process.returncode}.')
                        process_has_crashed = True
                        break
                    else:
                        # exited cleanly
                        log.info(f'Rank {global_rank} finished successfully.')
            if process_has_crashed or all_processes_finished:
                break
            time.sleep(0.1)
    except KeyboardInterrupt:
        print('Ctrl-C received; terminating training processes.')
        pass


def _print_process_exit_status(global_rank: int, process: subprocess.Popen):
    stdOutLabel = 'STDOUT'
    if process.stdout is None:
        output = None
    else:
        process.stdout.seek(0)
        output = process.stdout.read()

    if process.stderr is None:
        stderr = None
        stdOutLabel = 'logs'
    else:
        process.stderr.seek(0)
        stderr = process.stderr.read()
    exc = subprocess.CalledProcessError(
        process.returncode,
        cmd=process.args,
        output=output,
        stderr=stderr,
    )

    error_msg = [f'Global rank {global_rank} (PID {process.pid}) exited with code {process.returncode}']
    if output is not None:
        error_msg.extend([
            f'----------Begin global rank {global_rank} {stdOutLabel}----------',
            output,
            f'----------End global rank {global_rank} {stdOutLabel}----------',
        ])

    if stderr is not None:
        error_msg.extend([
            f'----------Begin global rank {global_rank} STDERR----------',
            exc.stderr,
            f'----------End global rank {global_rank} STDERR----------',
        ])
    print('\n'.join(error_msg))


def _cleanup_processes(processes: dict[int, subprocess.Popen]):
    for global_rank, process in processes.items():
        process.poll()
        if process.returncode is None:
            log.info('Killing global rank %s (PID %s) with SIGTERM', global_rank, process.pid)
            # Assuming that child processes correctly handle SIGTERM to cleanup any children
            try:
                os.kill(process.pid, signal.SIGTERM)
            except ProcessLookupError:
                pass

    current_time = datetime.datetime.now()

    try:
        print((
            f'Waiting up to {CLEANUP_TIMEOUT.seconds} seconds for all training processes to terminate. '
            'Press Ctrl-C to exit immediately.'
        ))
        while datetime.datetime.now() - current_time < CLEANUP_TIMEOUT:
            for process in processes.values():
                process.poll()
            if all(process.returncode is not None for process in processes.values()):
                break
            time.sleep(0.1)
    except KeyboardInterrupt:
        pass

    for global_rank, process in processes.items():
        process.poll()
        if process.returncode is None:
            log.warning(
                'Failed to kill global rank %s (PID %s) with SIGTERM; terminating with SIGKILL instead',
                global_rank,
                process.pid,
            )
            try:
                proc = psutil.Process(process.pid)
            except psutil.NoSuchProcess:
                pass
            else:
                # If using SIGKILL, manually kill all child processes, since the main training process
                # likely won't be able to intercept the signal and clean up its children.
                for psutil_proc in [proc, *proc.children(recursive=True)]:
                    try:
                        os.kill(psutil_proc.pid, signal.SIGKILL)
                    except ProcessLookupError:
                        pass
    for global_rank, process in processes.items():
        process.poll()
        if process.returncode is not None and process.returncode != 0:
            if -process.returncode in (signal.SIGKILL, signal.SIGTERM):
                # Negative return codes indicate the process was killed via a signal
                # If the launcher script killed the training process (which would happen via SIGKILL or SIGTERM),
                # then do not print the stack trace.
                continue
            # only print the processes that have actually crashed,
            # not the ones that were killed
            _print_process_exit_status(global_rank, process)


def _aggregate_process_returncode(processes: dict[int, subprocess.Popen]) -> int:
    for global_rank, process in processes.items():
        process.poll()
        if process.returncode is None:
            log.error('Global rank %s (PID %s) has still not exited; return exit code 1.', global_rank, process.pid)
            return 1
        if process.returncode != 0:
            log.error('Global rank %s (PID %s) exited with code %s', global_rank, process.pid, process.returncode)
            return process.returncode

    return 0


def main():
    """Entrypoint into the Composer CLI."""
    args = _parse_args()

    logging.basicConfig()
    log.setLevel(logging.INFO if args.verbose else logging.WARNING)

    processes = {}

    log_tmpdir = tempfile.TemporaryDirectory()
    if args.stdout is None:
        args.stdout = f'{log_tmpdir.name}/rank{{rank}}.stdout.txt'
    if args.stderr is None:
        args.stderr = f'{log_tmpdir.name}/rank{{rank}}.stderr.txt'

    # If running on the Mosaic platform, log all gpu ranks' stderr and stdout to Mosaic platform
    if os.environ.get(MOSAICML_PLATFORM_ENV_VAR, 'false').lower() == 'true' and str(
        os.environ.get(MOSAICML_LOG_DIR_ENV_VAR, 'false'),
    ).lower() != 'false' and os.environ.get(MOSAICML_GPU_LOG_FILE_PREFIX_ENV_VAR, 'false').lower() != 'false':
        log.info('Logging all GPU ranks to Mosaic AI Training.')
        log_file_format = f'{os.environ.get(MOSAICML_LOG_DIR_ENV_VAR)}/{os.environ.get(MOSAICML_GPU_LOG_FILE_PREFIX_ENV_VAR)}{{local_rank}}.txt'
        if args.stderr is not None or args.stdout is not None:
            log.info(
                'Logging to Mosaic AI Training. Ignoring provided stdout and stderr args. To use provided stdout and stderr, set MOSAICML_LOG_DIR=false.',
            )
        args.stdout = log_file_format
        args.stderr = None

    try:
        _launch_processes(
            nproc=args.nproc,
            world_size=args.world_size,
            base_rank=args.base_rank,
            node_rank=args.node_rank,
            master_addr=args.master_addr,
            master_port=args.master_port,
            module_mode=args.module_mode,
            command_mode=args.command_mode,
            stdout_file_format=args.stdout,
            stderr_file_format=args.stderr,
            training_script=args.training_script,
            training_script_args=args.training_script_args,
            processes=processes,
        )
        _monitor_processes(processes)
    except:
        # Print the exception first, then kill the training processes, since killing
        # may take up to CLEANUP_TIMEOUT seconds, and the user should know immediately
        # what failed. No need to re-raise the exception, as `aggregate_process_returncode`
        # will return an appropriate error code, which will cause the script to exit.
        traceback.print_exc()
        print('Killing training processes')
    finally:
        _cleanup_processes(processes)
        log_tmpdir.cleanup()
        return _aggregate_process_returncode(processes)


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Action states for the :class:`Profiler` that define whether or not events are being recorded to the trace file."""

from composer.utils import StringEnum

__all__ = ['ProfilerAction']


class ProfilerAction(StringEnum):
    """Action states for the :class:`Profiler` that define whether or not events are being recorded to the trace file.

    Attributes:
        SKIP: Do not record new events to the trace.  Any events started during ``ACTIVE`` or ``WARMUP`` will be recorded upon finish.
        WARMUP: Record all events to the trace `except` those requiring a warmup period to initialize data structures (e.g., :doc:`profiler`).
        ACTIVE: Record all events to the trace.
        ACTIVE_AND_SAVE: Record all events and save the trace at the end of the batch.
    """
    SKIP = 'skip'
    WARMUP = 'warmup'
    ACTIVE = 'active'
    ACTIVE_AND_SAVE = 'active_and_save'



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""OCI-Compatible object store."""

from __future__ import annotations

import concurrent.futures
import os
import pathlib
import uuid
from tempfile import TemporaryDirectory
from typing import Callable, Optional, Union

from composer.utils.import_helpers import MissingConditionalImportError
from composer.utils.object_store.object_store import ObjectStore

__all__ = ['OCIObjectStore']


def _reraise_oci_errors(uri: str, e: Exception):
    try:
        import oci
    except ImportError as e:
        raise MissingConditionalImportError(
            conda_package='oci',
            extra_deps_group='oci',
            conda_channel='conda-forge',
        ) from e

    # If it's an oci service error with code: ObjectNotFound or status 404
    if isinstance(e, oci.exceptions.ServiceError):
        if e.status == 404:  # type: ignore
            if e.code == 'ObjectNotFound':  # type: ignore
                raise FileNotFoundError(f'Object {uri} not found. {e.message}') from e  # type: ignore
            if e.code == 'BucketNotFound':  # type: ignore
                raise ValueError(f'Bucket specified in {uri} not found. {e.message}') from e  # type: ignore
            raise FileNotFoundError(f'Object {uri} not found with no error code. {e.message}') from e  # type: ignore

    # Client errors
    if isinstance(e, oci.exceptions.ClientError):
        raise ValueError(f'Error with using your OCI config file for uri {uri}') from e
    if isinstance(e, oci.exceptions.MultipartUploadError):
        raise ValueError(f'Error when uploading {uri} using OCI parallelized uploading') from e

    # Otherwise just raise the original error.
    raise e


class OCIObjectStore(ObjectStore):
    """Utility for uploading to and downloading from an OCI bucket.

    Args:
        bucket (str): The bucket name.
        prefix (str): A path prefix such as `folder/subfolder/` to prepend to object names. Defaults to ''.
    """

    def __init__(
        self,
        bucket: str,
        prefix: str = '',
    ) -> None:
        try:
            import oci
        except ImportError as e:
            raise MissingConditionalImportError(
                conda_package='oci',
                extra_deps_group='oci',
                conda_channel='conda-forge',
            ) from e

        # Format paths
        self.bucket = bucket.strip('/')
        self.prefix = prefix.strip('/')
        if self.prefix != '':
            self.prefix += '/'

        try:
            if 'OCI_CONFIG_FILE' in os.environ:
                config = oci.config.from_file(os.environ['OCI_CONFIG_FILE'])
            else:
                config = oci.config.from_file()

            self.client = oci.object_storage.ObjectStorageClient(
                config=config,
                retry_strategy=oci.retry.DEFAULT_RETRY_STRATEGY,
            )
        except Exception as e:
            _reraise_oci_errors(self.get_uri(object_name=''), e)

        self.namespace = self.client.get_namespace().data  # pyright: ignore[reportOptionalMemberAccess]
        self.upload_manager = oci.object_storage.UploadManager(self.client)

    def get_uri(self, object_name: str) -> str:
        return f'oci://{self.bucket}/{object_name}'

    def get_object_size(self, object_name: str) -> int:
        try:
            response = self.client.get_object(
                namespace_name=self.namespace,
                bucket_name=self.bucket,
                object_name=object_name,
            )
        except Exception as e:
            _reraise_oci_errors(self.get_uri(object_name), e)

        if response.status == 200:  # pyright: ignore[reportUnboundVariable, reportOptionalMemberAccess]
            data = response.data  # pyright: ignore[reportUnboundVariable, reportOptionalMemberAccess]
            return int(data.headers['Content-Length'])
        else:
            status = response.status  # pyright: ignore[reportUnboundVariable, reportOptionalMemberAccess]
            raise ValueError(f'OCI get_object was not successful with a {status} status code.')

    def upload_object(
        self,
        object_name: str,
        filename: Union[str, pathlib.Path],
        callback: Optional[Callable[[int, int], None]] = None,
    ):
        del callback
        try:
            self.upload_manager.upload_file(
                namespace_name=self.namespace,
                bucket_name=self.bucket,
                object_name=object_name,
                file_path=filename,
            )

        except Exception as e:
            _reraise_oci_errors(self.get_uri(object_name), e)

    def _download_part(self, object_name, filename, start_byte, end_byte, part_number):
        range_header = f'bytes={start_byte}-{end_byte}'
        tmp_part_path = os.path.join(filename, f'part-{part_number}-{uuid.uuid4()}.tmp')
        response = self.client.get_object(
            namespace_name=self.namespace,
            bucket_name=self.bucket,
            object_name=object_name,
            range=range_header,
        )
        with open(tmp_part_path, 'wb') as f:
            f.write(response.data.content)  # pyright: ignore[reportOptionalMemberAccess]
        return part_number, tmp_part_path

    def download_object(
        self,
        object_name: str,
        filename: Union[str, pathlib.Path],
        overwrite: bool = False,
        callback: Optional[Callable[[int, int], None]] = None,
        min_part_size: int = 128000000,
        num_parts: int = 10,
    ):
        del callback
        if os.path.exists(filename) and not overwrite:
            raise FileExistsError(f'The file at {filename} already exists and overwrite is set to False')

        dirname = os.path.dirname(filename)
        if dirname:
            os.makedirs(dirname, exist_ok=True)

        # Get the size of the object
        object_size = 0
        try:
            head_object_response = self.client.head_object(self.namespace, self.bucket, object_name)
            object_size = int(
                head_object_response.headers['content-length'],  # pyright: ignore[reportOptionalMemberAccess]
            )
        except Exception as e:
            _reraise_oci_errors(self.get_uri(object_name), e)

        # Calculate the part sizes
        num_parts_from_size = max(object_size // min_part_size, 1)
        num_parts = min(num_parts, num_parts_from_size)
        base_part_size, remainder = divmod(object_size, num_parts)
        part_sizes = [base_part_size] * num_parts
        for i in range(remainder):
            part_sizes[i] += 1
        part_sizes = [part_size for part_size in part_sizes if part_size > 0]

        with TemporaryDirectory(dir=dirname, prefix=f'{str(filename)}') as temp_dir:
            parts = []
            try:
                # Download parts in parallel
                with concurrent.futures.ThreadPoolExecutor() as executor:
                    futures = []
                    start_byte = 0
                    for i, part_size in enumerate(part_sizes):
                        end_byte = start_byte + part_size - 1
                        futures.append(
                            executor.submit(self._download_part, object_name, temp_dir, start_byte, end_byte, i),
                        )
                        start_byte = end_byte + 1

                    for future in concurrent.futures.as_completed(futures):
                        parts.append(future.result())
                    parts = sorted(parts, key=lambda x: x[0])
            except Exception as e:
                _reraise_oci_errors(self.get_uri(object_name), e)

            # Combine parts
            tmp_path = os.path.join(temp_dir, f'{str(filename)}-{uuid.uuid4()}.tmp')
            with open(tmp_path, 'wb') as outfile:
                for i, part_file_name in parts:
                    with open(part_file_name, 'rb') as infile:
                        outfile.write(infile.read())

            if overwrite:
                os.replace(tmp_path, filename)
            else:
                os.rename(tmp_path, filename)

    def list_objects(self, prefix: Optional[str] = None) -> list[str]:
        if prefix is None:
            prefix = ''

        if self.prefix:
            prefix = f'{self.prefix}{prefix}'

        object_names = []
        next_start_with = None
        response_complete = False
        try:
            while not response_complete:
                response = self.client.list_objects(
                    namespace_name=self.namespace,
                    bucket_name=self.bucket,
                    prefix=prefix,
                    start=next_start_with,
                ).data  # pyright: ignore[reportOptionalMemberAccess]
                object_names.extend([obj.name for obj in response.objects])
                next_start_with = response.next_start_with
                if not next_start_with:
                    response_complete = True
        except Exception as e:
            _reraise_oci_errors(self.get_uri(prefix), e)

        return object_names



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Core code for sequence length warmup."""

import logging
import textwrap
from typing import Mapping, Optional

import torch
import torch.utils.data

from composer.core import Algorithm, Batch, Event, State, TimeUnit, get_precision_context
from composer.loggers import Logger
from composer.models import HuggingFaceModel
from composer.utils import dist, ensure_tuple

log = logging.getLogger(__name__)

__all__ = ['SeqLengthWarmup', 'set_batch_sequence_length']


def set_batch_sequence_length(
    batch: dict[str, torch.Tensor],
    curr_seq_len: int,
    truncate: bool = True,
    preserve_end_of_sequence: bool = False,
) -> Batch:
    """Set the sequence length of a batch.

    Changes the sequence length of all tensors in the provided dictionary
    to ``curr_seq_len`` by either truncating the tensors (``truncate=True``)
    or reshaping the tensors to create new examples from the extra tokens
    (``truncate=False``).

    .. note::

        The schedule for ``curr_seq_len`` over training time should be managed
        outside of this function.

    .. note::

        Variable input lengths can create CUDA OOM errors. To avoid this,
        we follow the `PyTorch notes <https://pytorch.org/tutorials/recipes/recipes/tuning_guide.html#pre-allocate-memory-in-case-of-variable-input-length>`_
        and pre-allocate the memory with a blank forward and backward pass.

    Args:
        batch (dict[str, Tensor]): The input batch to the model, must be a dictionary.
        curr_seq_length (int): The desired sequence length to apply.
        truncate (bool, optional): Truncate sequences early, or reshape tensors to create
            new examples out of the extra tokens. Default: ``True``.
        preserve_end_of_sequence (bool, optional): Preserve the end-of-sequence of the batch when
            truncating. Useful when input formats include a unique end-of-sequence token.
            Ignored if ``truncate=False``. Default: ``False``.
            E.g., if ``batch["input_ids"]`` is ``[[10, 11, 12, 13, 14, 15]]``
            and ``curr_seq_length=3``, ``"input_ids"`` in the returned batch would be
            ``[[10, 11, 12]]`` with ``preserve_end_of_sequence=False`` and would be
            ``[[10, 11, 15]]`` with ``preserve_end_of_sequence=True``. This behavior applies to any
            batch tensor with 2 or more dimensions.

    Returns:
        dict[str, Tensor]: a Mapping of input tensors to the model,
            where all tensors have curr_seq_len in the second dimension.

    Example:

    .. code-block::

        import composer.functional as cf

        for epoch in range(num_epochs):
            for X, y in train_loader:
                X = cf.set_batch_sequence_length(X, sequence_length)
                y_hat = model(X)
                loss = loss_fn(y_hat, y)
    """

    assert isinstance(batch, Mapping)

    # This should act like a no-op if curr_seq_len isn't shorter than the batch's sequence length
    batch_seq_len = 0
    # The batch sequence length is assumed to be the shape[1] dimension of any non-1D batch tensor
    for batch_tensor in batch.values():
        tensor_shape = batch_tensor.shape
        if len(tensor_shape) > 1:
            batch_seq_len = tensor_shape[1]
            break
    if curr_seq_len >= batch_seq_len:
        return batch

    if truncate:
        # Truncate, but preserve end-of-sequence tokens
        if preserve_end_of_sequence:
            if 'attention_mask' not in batch:
                raise ValueError(
                    'Sequence Length Warmup requires that the batch has "attention_mask" when using ``preserve_end_of_sequence=True``.',
                )
            r_idx = torch.arange(batch['attention_mask'].shape[0])
            # eos_idx should point to the final token index for each batch sample
            eos_idx = batch['attention_mask'].sum(dim=1).long() - 1
            # eos_idx_truncated is the same thing, after truncation is applied
            eos_idx_truncated = eos_idx.clamp(max=curr_seq_len - 1)

            for k in batch.keys():
                if batch[k].ndim < 2:
                    raise ValueError(
                        f'Sequence Length Warmup requires that all tensors are sequence-shaped when ``truncate=True``. '
                        f'Tensor "{k}" has shape {batch[k].shape}.',
                    )
                eos_value = batch[k][r_idx, eos_idx]
                batch[k] = batch[k][:, :curr_seq_len].contiguous()
                batch[k][r_idx, eos_idx_truncated] = eos_value

        else:
            for k in batch.keys():
                if batch[k].ndim < 2:
                    raise ValueError(
                        f'Sequence Length Warmup requires that all tensors are sequence-shaped when ``truncate=True``. '
                        f'Tensor "{k}" has shape {batch[k].shape}.',
                    )
                batch[k] = batch[k][:, :curr_seq_len].contiguous()

    else:
        if 'input_ids' not in batch:
            raise ValueError(
                'Sequence Length Warmup requires that the batch has "input_ids" when using ``truncate=False``.',
            )
        input_ids_shape = batch['input_ids'].shape
        # ensure new tensor shape is divisible by curr_seq_len
        input_ids = batch['input_ids'].view(-1)
        tensor_len = (input_ids.shape[0] // curr_seq_len) * curr_seq_len

        input_ids = input_ids[:tensor_len]
        input_ids = input_ids.view(-1, curr_seq_len)
        batch['input_ids'] = input_ids

        for k, v in batch.items():
            if k == 'input_ids':
                continue
            if v.shape != input_ids_shape:
                raise ValueError(
                    f'When using ``truncate=False``, Sequence Length Warmup only supports batches where all tensors have the same shape. '
                    f'Tensor "{k}" has shape {v.shape} but should have shape {input_ids_shape}.',
                )
            v = v.view(-1)
            v = v[:tensor_len]
            batch[k] = v.view(-1, curr_seq_len)

    return batch


class SeqLengthWarmup(Algorithm):
    """Progressively increases the sequence length during training.

    Changes the sequence length of all tensors in the input batch. The
    sequence length increases from ``min_seq_length`` to ``max_seq_length``
    in steps of ``step_size`` during the first ``duration`` fraction of
    training.

    The sequence length is then kept at ``max_seq_length``
    for the rest of training.

    Tensors are either truncated (``truncate=True``) or reshaped to
    create new examples from the extra tokens (``truncate=False``).

    This algorithm runs on :attr:`.Event.AFTER_DATALOADER` to modify
    the sequence length of a batch of data after the model and data have been moved to
    accelerators.

    .. note::

        ``step_size`` should be a `multiple of eight <https://developer.nvidia.com/blog/optimizing-gpu-performance-tensor-cores/>`_ for
        optimal throughput on NVIDIA GPUs.

    .. note::

        Variable input lengths can create CUDA OOM errors. To avoid this,
        we follow the `PyTorch notes <https://pytorch.org/tutorials/recipes/recipes/tuning_guide.html#pre-allocate-memory-in-case-of-variable-input-length>`_
        and pre-allocate the memory with a blank forward and backward pass.

    See the :doc:`Method Card </method_cards/seq_length_warmup>` for more details.

    Example:

    .. code-block::

        from composer.algorithms import SeqLengthWarmup
        from composer import Trainer

        seq_length_warmup = SeqLengthWarmup(duration=0.5,
                                            min_seq_length=8,
                                            max_seq_length=1024,
                                            step_size=8,
                                            truncate=True,
                                            preserve_end_of_sequence=False)

        trainer = Trainer(model=model,
                          train_dataloader=train_dataloader,
                          max_duration="1ep",
                          algorithms=[seq_length_warmup])

    Args:
        duration (float, optional): Fraction of total training for sequential length
            learning. Default = ``0.3``.
        min_seq_length (int, optional): Minimum sequence length to start the warmup.
            Default = ``8``.
        max_seq_length (int, optional): Maximum sequence length to stop the warmup.
            Default = ``1024``.
        step_size (int, optional): Step size of sequence length. Default = ``8``.
        truncate (bool, optional): Truncate sequences early, or reshape tensors to create
            new examples out of the extra tokens. Default: ``True``.
        preserve_end_of_sequence (bool, optional): Preserve the end-of-sequence of the batch when
            truncating. Useful when input formats include a unique end-of-sequence token.
            Ignored if ``truncate=False``. Default: ``False``.
            E.g., if ``batch["input_ids"]`` is ``[[10, 11, 12, 13, 14, 15]]``
            and ``curr_seq_length=3``, ``"input_ids"`` in the returned batch would be
            ``[[10, 11, 12]]`` with ``preserve_end_of_sequence=False`` and would be
            ``[[10, 11, 15]]`` with ``preserve_end_of_sequence=True``. This behavior applies to any
            batch tensor with 2 or more dimensions.
    """

    def __init__(
        self,
        duration: float = 0.3,
        min_seq_length: int = 8,
        max_seq_length: int = 1024,
        step_size: int = 8,
        truncate: bool = True,
        preserve_end_of_sequence: bool = False,
    ):
        self.duration = duration
        self.min_seq_length = min_seq_length
        self.max_seq_length = max_seq_length
        self.step_size = step_size
        self.truncate = truncate
        self.preserve_end_of_sequence = preserve_end_of_sequence

        if self.duration < 0 or self.duration > 1:
            raise ValueError(f'Duration must be between 0 and 1, got: {self.duration}')

        if self.max_seq_length < self.min_seq_length:
            raise ValueError(
                f'max_seq_length={self.max_seq_length} must be '
                f'greater than min_seq_length={self.min_seq_length}',
            )
        self._activated = False
        self._original_model = None

    def _activate_model(self, state: State, logger: Logger) -> None:
        """Does a forward and a backward pass on a dummy input.

        The purpose of activating the model is to prevent OOMs. This happens two ways.

        First, this prevents GPU memory from being reallocated when the sequence
        length increases.

        Second, it detects if the batch*max_sequence_length size will cause an OOM and
        decreases state.device_train_microbatch_size accordingly. This logic mirrors the
        ``device_train_microbatch_size='auto'`` logic in :class:`.Trainer`.
        """

        assert self._original_model is not None, 'original model should be set on Event.INIT'

        try:
            # Both PyTorch and FFCV dataloaders define a `batch_size` attribute
            # This exception would mainly be raised if the user is passing in a custom
            # iterable
            per_gpu_macrobatch = getattr(state.dataloader, 'batch_size')
        except AttributeError as e:
            raise AttributeError(
                'Sequence Length Warmup requires the `state.dataloader` to have a `batch_size` attribute.',
            ) from e
        if per_gpu_macrobatch is None:
            raise RuntimeError('Sequence Length Warmup algorithm requires constant batch size.')

        # truncate all sequence-shaped tensors to the max sequence length
        batch_clone = {k: torch.clone(v) for k, v in state.batch.items()}
        for k, v in batch_clone.items():
            if v.ndim < 2:
                raise ValueError(
                    f'Sequence Length Warmup requires that all tensors are sequence-shaped. '
                    f'Tensor "{k}" has shape {v.shape}.',
                )
            batch_clone[k] = v[:, :self.max_seq_length].contiguous()

        # In-line to avoid circular dependency
        from composer.trainer.trainer import _adjust_device_train_microbatch_size, _is_cuda_oom

        # This loop tries to do a forward/backward pass using the current microbatch size.
        # If it hits an OOM error, it halves `state.device_train_microbatch_size` and tries again
        # until it succeeds.
        while True:
            model_inputs = {k: v[:state.device_train_microbatch_size] for k, v in batch_clone.items()}

            found_cuda_oom = 0  # int since bool BOR not supported on all torch.distributed backends
            try:
                # Start by running a forward and backward pass
                # of the maximum sequence length to allocate cache.
                with get_precision_context(state.precision, state.precision_config):
                    outputs = state.model.forward(model_inputs)
                    loss = self._original_model.loss(outputs, model_inputs)

                # Check if other ranks OOMed after forward pass when using auto microbatching. This may
                # happen when close to memory limit or with uneven memory usage across ranks
                if state.auto_microbatching:
                    # Check if any other rank hit an OOM
                    found_cuda_oom_tensor = state.device.tensor_to_device(torch.tensor([0], dtype=torch.uint8))
                    dist.all_reduce(found_cuda_oom_tensor, reduce_operation='MAX')
                    found_cuda_oom = found_cuda_oom_tensor.item()
                    # Signal current rank is still in batch
                    all_ranks_finished_tensor = state.device.tensor_to_device(torch.tensor([0], dtype=torch.uint8))
                    dist.all_reduce(all_ranks_finished_tensor, reduce_operation='MIN')

                    if found_cuda_oom == 1:
                        raise RuntimeError('CUDA out of memory encountered on a different rank')

                # Since use_grad_scaling is in the Trainer, and we
                # don't care about the loss values, skip scaling
                for loss_item in ensure_tuple(loss):
                    loss_item.backward()

                # Zero any gradients created by the backward pass
                for optimizer in state.optimizers:
                    optimizer.zero_grad()

            # This error/state.device_train_microbatch_size handling mimics the logic in trainer._train_batch().
            except RuntimeError as e:
                if state.auto_microbatching and _is_cuda_oom(e):
                    log.debug((f"Rank {dist.get_global_rank()} OOM'd."))
                    found_cuda_oom = 1
                elif state.auto_microbatching and ('cuda' in str(e).lower() or 'c10' in str(e).lower()):
                    raise ValueError(
                        textwrap.dedent(
                            'Encountered non-addressable cuda error while using auto microbatching. '
                            'If this repeatedly occurs, set `device_train_microbatch_size` manually.',
                        ),
                    ) from e
                else:
                    raise

            if state.auto_microbatching:
                all_ranks_finished = False
                while not all_ranks_finished:
                    # Propagate across all ranks if any rank hit CUDA OOM
                    found_cuda_oom_tensor = state.device.tensor_to_device(
                        torch.tensor([found_cuda_oom], dtype=torch.uint8),
                    )
                    dist.all_reduce(found_cuda_oom_tensor, reduce_operation='MAX')
                    found_cuda_oom = found_cuda_oom_tensor.item()
                    # Check if any rank is still not done with the batch. This may happen if only a
                    # subset of ranks OOM, leaving some batches still in the forward pass
                    all_ranks_finished_tensor = state.device.tensor_to_device(torch.tensor([1], dtype=torch.uint8))
                    dist.all_reduce(all_ranks_finished_tensor, reduce_operation='MIN')
                    all_ranks_finished = all_ranks_finished_tensor.item() == 1
                if found_cuda_oom == 1:
                    _adjust_device_train_microbatch_size(state)
                    # Skip return and rerun after handling oom
                    continue
            # Activate and return if we've completed without OOMing.
            self._activated = True
            return

    def match(self, event: Event, state: State) -> bool:
        return (event == Event.INIT and self._original_model is None) or event == Event.AFTER_DATALOADER

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        if event == Event.INIT:
            if not isinstance(state.model, HuggingFaceModel):
                raise RuntimeError(
                    textwrap.dedent(
                        f"""\
                    {type(self).__name__} requires state.model to be of type {HuggingFaceModel.__name__}, not of type {type(state.model)}""",
                    ),
                )

            self._original_model = state.model
            return

        assert state.dataloader is not None, 'dataloader should be set on AFTER_DATALOADER'
        assert state.max_duration is not None, 'max_duration should be set on AFTER_DATALOADER'

        # in order to avoid OOMs, we do a forward and a backward pass on a dummy input.
        if not self._activated:
            self._activate_model(state, logger)

        if state.max_duration.unit == TimeUnit.EPOCH:
            if state.dataloader_len is None:
                raise RuntimeError('Sequential Length Warmup requires the dataloader to be sized.')
            num_optimization_steps = int(state.dataloader_len) * state.max_duration.value
        elif state.max_duration.unit == TimeUnit.BATCH:
            num_optimization_steps = state.max_duration.value
        else:
            raise NotImplementedError(
                textwrap.dedent(
                    """\
                    To use sequential length warmup, the max_duration must be in epochs or batches.
                    Specifying the `max_duration` in tokens or samples for use with sequential
                    length warmup will be supported in a future Composer release. See
                    https://github.com/mosaicml/composer/issues/226.""",
                ),
            )
        num_warmup_steps = int(num_optimization_steps * self.duration)  # in batches

        # assume the full sequence length is the unaltered sequence length
        num_update_steps = (self.max_seq_length - self.min_seq_length) // self.step_size
        update_every_n_steps = num_warmup_steps // num_update_steps

        curr_seq_len = self.step_size * (int(state.timestamp.batch) // update_every_n_steps) + self.min_seq_length
        curr_seq_len = max(curr_seq_len, self.min_seq_length)
        curr_seq_len = min(curr_seq_len, self.max_seq_length)

        state.batch = set_batch_sequence_length(state.batch, curr_seq_len, self.truncate, self.preserve_end_of_sequence)

        batch_size = state.batch['input_ids'].shape[0]
        logger.log_metrics({
            'seq_length_warmup/curr_seq_len': curr_seq_len,
            'seq_length_warmup/curr_bs': batch_size,
        })



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Utility functions for torch profiler."""

import importlib.util
import logging
from base64 import b64encode
from os import remove
from tempfile import NamedTemporaryFile
from typing import Any, Optional, Union

import numpy as np
import torch
import torch.cuda
from packaging import version
from torch.profiler.profiler import profile as TorchProfile

log = logging.getLogger(__name__)


def export_memory_timeline_html(
    prof: TorchProfile,
    path: str,
    device: Optional[str] = None,
    figsize=(20, 12),
    title=None,
    yxis_step_size: float = 1.0,
    return_fig: bool = False,
) -> Optional[Union[None, Any]]:
    """Exports a memory timeline to an HTML file. Similar to the PyTorch plotting function, but with adjusted axis tickers and grids."""
    if version.parse(torch.__version__) <= version.parse('2.1.0.dev'):
        log.warning('export_memory_timeline_html failed because memory timeline is supported after PyTorch 2.1.0.')
        return

    from torch.profiler._memory_profiler import _CATEGORY_TO_COLORS, _CATEGORY_TO_INDEX, MemoryProfileTimeline

    # Default to device 0, if unset. Fallback on cpu.
    if device is None and prof.use_device and prof.use_device != 'cuda':
        device = prof.use_device + ':0'

    if device is None:
        device = 'cuda:0' if torch.cuda.is_available() else 'cpu'

    # Construct the memory timeline plot data
    mem_tl = MemoryProfileTimeline(prof._memory_profile())

    # Check if user has matplotlib installed, return gracefully if not.
    matplotlib_spec = importlib.util.find_spec('matplotlib')
    if matplotlib_spec is None:
        log.warning('export_memory_timeline_html failed because matplotlib was not found.')
        return
    import matplotlib.pyplot as plt

    mt = mem_tl._coalesce_timeline(device)
    times, sizes = np.array(mt[0]), np.array(mt[1])
    stacked = np.cumsum(sizes, axis=1) / 1024**3
    max_memory_allocated = torch.cuda.max_memory_allocated()
    max_memory_reserved = torch.cuda.max_memory_reserved()

    # Plot memory timeline as stacked data
    fig = plt.figure(figsize=figsize, dpi=80)
    axes = fig.gca()
    for category, color in _CATEGORY_TO_COLORS.items():
        i = _CATEGORY_TO_INDEX[category]
        axes.fill_between(times / 1e3, stacked[:, i], stacked[:, i + 1], color=color, alpha=0.7)
    fig.legend(['Unknown' if i is None else i.name for i in _CATEGORY_TO_COLORS])
    axes.set_xlabel('Time (us)')
    axes.set_ylabel('Memory (GB)')
    _, end = axes.get_ylim()
    axes.grid(True)
    axes.set_yticks(np.arange(0, end, yxis_step_size))
    title = '\n\n'.join(([title] if title else []) + [
        f'Max memory allocated: {max_memory_allocated/(10**9):.2f} GB \n'
        f'Max memory reserved: {max_memory_reserved/(10**9):.2f} GB',
    ])
    axes.set_title(title)

    if return_fig:
        return fig

    # Embed the memory timeline image into the HTML file
    tmpfile = NamedTemporaryFile('wb', suffix='.png', delete=False)
    tmpfile.close()
    fig.savefig(tmpfile.name, format='png')

    with open(tmpfile.name, 'rb') as tmp:
        encoded = b64encode(tmp.read()).decode('utf-8')
        html = f"""<html>
                <head><meta charset="utf-8" /><title>GPU Memory Timeline HTML</title></head>
                <body>
                <img src='data:image/png;base64,{encoded}'>
                </body>
                </html>"""

        with open(path, 'w') as f:
            f.write(html)
    log.debug('Memory timeline exported to %s.', path)
    remove(tmpfile.name)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""A wrapper class that converts 🤗 Transformers models to composer models"""

from __future__ import annotations

import copy
import inspect
import json
import logging
import os
import random
import string
import tempfile
import textwrap
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Mapping, Optional, Type, Union

import torch
from torchmetrics import Metric

from composer.devices import DeviceCPU
from composer.models.base import ComposerModel
from composer.utils import MissingConditionalImportError, dist, get_file, import_object, is_model_fsdp, safe_torch_load

try:
    from peft import PeftModel, get_peft_model
    peft_installed = True
except:
    peft_installed = False

if TYPE_CHECKING:
    import transformers
    from peft import PeftConfig, PeftModel
    from transformers import PretrainedConfig
    from transformers.models.auto.auto_factory import _BaseAutoModelClass

log = logging.getLogger(__name__)

__all__ = ['HuggingFaceModel', 'peft_installed']


class HuggingFaceModel(ComposerModel):
    """
    A wrapper class that converts 🤗 Transformers models to composer models.

    Args:
        model (Union[transformers.PreTrainedModel, peft.PeftModel)): A 🤗 Transformers model or a PEFT model.
        tokenizer (transformers.PreTrainedTokenizer, optional): The tokenizer used to prepare the dataset. Default ``None``.

            .. note:: If the tokenizer is provided, its config will be saved in the composer checkpoint, and it can be reloaded
                using :meth:`HuggingFaceModel.hf_from_composer_checkpoint`. If the tokenizer is not provided here, it will not be saved in the composer checkpoint.
        use_logits (bool, optional): If True, the model's output logits will be used to calculate validation metrics. Else, metrics will be inferred from the HuggingFaceModel directly. Default: ``False``
        metrics (list[Metric], optional): list of torchmetrics to apply to the output of `eval_forward` during training. If ``eval_metrics`` is ``None``, these will also be used as ``eval_metrics``.  Default: ``None``.
        eval_metrics (list[Metric], optional): list of torchmetrics to compute on the eval_dataloader, or be accessible to :class:`Evaluator`s. Default: ``None``.
        shift_labels (bool, optional): If True, the batch's labels will be shifted before being used to calculate metrics. This should be set to true for CausalLM models and false otherwise. If not specified, `shift_labels` will be set automatically based on the model class name. Default: ``None``.
        allow_embedding_resizing (bool, optional): If True, the model's embeddings will be automatically resized when they are smaller than the tokenizer vocab size. Default: ``False``.
        peft_config (PeftConfig, optional): Optional PEFT config to apply to the model. If provided, the model will be converted to a PEFT model. Only LoRA is currently supported.
        should_save_peft_only (bool, optional): If True _and_ PEFT is active, the state dict will only contain the PEFT weights, not the frozen base model weights.

        .. note:: To ensure correct behavior, set `shift_labels` manually if using a custom model (i.e., if `model` is not
        an instance of a registered 🤗 Transformers class).
    .. warning:: This wrapper is designed to work with 🤗 datasets that define a `labels` column.

    Example:

    .. testcode::

        import transformers
        from composer.models import HuggingFaceModel

        hf_model = transformers.AutoModelForSequenceClassification.from_pretrained('google-bert/bert-base-uncased', num_labels=2)
        hf_tokenizer = transformers.AutoTokenizer.from_pretrained('google-bert/bert-base-uncased')
        model = HuggingFaceModel(hf_model, hf_tokenizer)
    """

    def __init__(
        self,
        model: Union[transformers.PreTrainedModel, 'PeftModel'],
        tokenizer: Optional[Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast]] = None,
        use_logits: Optional[bool] = False,
        metrics: Optional[list[Metric]] = None,
        eval_metrics: Optional[list[Metric]] = None,
        shift_labels: Optional[bool] = None,
        allow_embedding_resizing: bool = False,
        peft_config: Optional['PeftConfig'] = None,
        should_save_peft_only: bool = True,
    ) -> None:
        try:
            import transformers
            del transformers  # unused
        except ImportError as e:
            raise MissingConditionalImportError(
                extra_deps_group='nlp',
                conda_package='transformers',
                conda_channel='conda-forge',
            ) from e

        if peft_config is not None:
            if not peft_installed:
                raise MissingConditionalImportError(
                    extra_deps_group='peft',
                    conda_package='peft',
                    conda_channel='conda-forge',
                )

        if peft_config is not None:
            # Hugging Face requires the peft type and task type to be upper case, so we do that here
            # https://github.com/huggingface/peft/blob/ebbff4023ad276cbcb2466fd7e99be7d3ae0ae11/src/peft/utils/peft_types.py#L22-L51
            if isinstance(peft_config.peft_type, str):
                peft_config.peft_type = peft_config.peft_type.upper()
            if isinstance(peft_config.task_type, str):
                peft_config.task_type = peft_config.task_type.upper()

            if peft_config.peft_type != 'LORA':
                raise ValueError(
                    f'PEFT type {peft_config.peft_type} is not supported by HuggingFaceModel. Only LORA is supported.',
                )

        super().__init__()
        self.model = model
        self.config: PretrainedConfig = model.config
        self.model_forward_args = self._get_model_forward_args()
        self.tokenizer = tokenizer
        self.should_save_peft_only = should_save_peft_only
        self.use_logits = use_logits
        self.labels: Optional[torch.Tensor] = None  # set in eval_forward() if exists
        self.dummy_forward_called = False  # Used to make FSDP generate work, see generate function for more details
        self.train_metrics: Optional[dict] = self._get_metric_dict(metrics) if metrics is not None else None
        self.val_metrics: Optional[dict] = self._get_metric_dict(
            eval_metrics,
        ) if eval_metrics is not None else copy.deepcopy(
            self.train_metrics,
        )

        is_causal_lm = _is_registered_causal_lm(self.model)
        self.shift_labels = is_causal_lm if shift_labels is None else shift_labels

        self._check_tokenizer_and_maybe_resize_embeddings(allow_embedding_resizing)

        if is_causal_lm and not self.shift_labels:
            log.warning(
                'The shift_labels argument was set to False but the model is an instance of a'
                ' HuggingFace Causal LM. This may lead to incorrect behavior.',
            )
            # Note: No warning if shift_labels and not is_causal_lm, since the model may simply be a custom class.

        if peft_config is not None:
            self.model = _maybe_get_peft_model(peft_config, self.model)

        self.using_peft = isinstance(self.model, PeftModel) if peft_installed else False

    def _check_tokenizer_and_maybe_resize_embeddings(self, allow_embedding_resizing: bool) -> None:
        if self.tokenizer is None:
            log.warning(
                'The tokenizer was not provided. This means the tokenizer config will not be saved in the checkpoint.',
            )

        if self.tokenizer is not None and self.config.vocab_size < len(self.tokenizer):
            if allow_embedding_resizing:
                # when the embedding size is smaller than the tokenizer vocab size,
                # the embeddings should get resized to match the tokenizer vocab size
                log.warning(
                    f'The number of tokens in the tokenizer is greater than the number of tokens in the model.'
                    f' This would cause an error during training.'
                    f' Resizing the model embeddings to {len(self.tokenizer)} from {self.config.vocab_size}.',
                )
                self.model.resize_token_embeddings(len(self.tokenizer))
            else:
                raise ValueError(
                    f'The number of tokens in the tokenizer is greater than the number of tokens in the model.'
                    f' This would cause an error during training.'
                    f' You can resize the model embeddings to {len(self.tokenizer)} from {self.config.vocab_size}'
                    f' by calling `model.resize_token_embeddings(len(tokenizer))` before calling the `HuggingFaceModel`'
                    f' constructor, or pass `allow_embedding_resizing=True` to have it done automatically.',
                )
        elif self.tokenizer is not None and self.config.vocab_size > len(self.tokenizer):
            # when the embedding size is greater than the tokenizer vocab size,
            # the embeddings do not _need_ to be resized to match the tokenizer vocab size,
            # and should be done by the user if desired
            log.info(
                f'The number of tokens in the tokenizer is less than the number of tokens in the model.'
                f' You may want to resize the model embeddings to {len(self.tokenizer)} from {self.config.vocab_size}'
                f' by calling `model.resize_token_embeddings(len(tokenizer))` before calling the `HuggingFaceModel`'
                f' constructor. The vocab size is sometimes intentionally set to a multiple of 32 or 64 to improve'
                f' performance.',
            )

    def _get_metric_dict(self, metrics: list[Metric]) -> dict[str, Metric]:
        """Returns a dictionary of metrics keyed by their class name."""
        return {metric.__class__.__name__: metric for metric in metrics}

    def _get_model_forward_args(self) -> set[str]:
        """Returns the arguments to the model's forward function."""
        model_forward_args = inspect.signature(maybe_get_underlying_model(self.model).forward).parameters.keys()

        if not model_forward_args:
            raise ValueError('Could not determine the forward arguments of the model. Please open a GitHub issue.')

        model_forward_args = set(model_forward_args)

        return model_forward_args

    def state_dict(self, *args, **kwargs) -> dict[str, Any]:
        """Returns the state dict of the model."""
        full_state_dict = super().state_dict(*args, **kwargs)

        if self.using_peft and self.should_save_peft_only:
            active_adapter = self.model.active_adapter
            assert isinstance(active_adapter, str)
            full_state_dict = filter_state_dict_peft(
                full_state_dict,
                self.model.peft_config[active_adapter],
                adapter_name='default',
                remove_adapter_names=False,
            )

        return full_state_dict

    @staticmethod
    def load_huggingface_tokenizer_from_saved_state(
        hf_state: dict[str, Any],
        trust_remote_code: bool = False,
        tokenizer_save_dir: Optional[str] = None,
    ) -> Optional[transformers.PreTrainedTokenizer | transformers.PreTrainedTokenizerFast]:
        """A helper function that loads a HuggingFace tokenizer from a loaded in hf state.

        Args:
            hf_state (dict[str, Any]): HF state loaded from a Composer checkpoint.
            trust_remote_code (bool, optional): Whether to trust the remote code when loading the tokenizer. Defaults to False.
            tokenizer_save_dir (Optional[str], optional): If specified, where to save the tokenizer files to locally. If not specified,
                a folder with a unique suffix will be saved in the current working directory. Defaults to None.

        Returns:
            Optional[transformers.PreTrainedTokenizer | transformers.PreTrainedTokenizerFast]: The loaded HuggingFace tokenizer
        """
        try:
            import transformers
        except ImportError as e:
            raise MissingConditionalImportError(
                extra_deps_group='nlp',
                conda_package='transformers',
                conda_channel='conda-forge',
            ) from e
        hf_tokenizer = None
        hf_tokenizer_state = hf_state['tokenizer']
        if hf_tokenizer_state != {}:
            if tokenizer_save_dir is None:
                unique_suffix = ''.join(random.choices(string.ascii_letters + string.digits, k=6))
                tokenizer_save_dir = os.path.join(os.getcwd(), f'tokenizer-save-dir-{unique_suffix}')
            os.makedirs(tokenizer_save_dir, exist_ok=True)

            for filename, saved_content in hf_tokenizer_state.items():
                # This cannot be a temporary directory because huggingface relies on the slow tokenizer file
                # being persistent on disk

                # For backwards compatibility, check if the filename already has the file extension
                if filename.endswith(saved_content['file_extension']):
                    tokenizer_file_name = filename
                else:
                    tokenizer_file_name = filename + saved_content['file_extension']

                tokenizer_file_path = Path(tokenizer_save_dir) / tokenizer_file_name
                if saved_content['file_extension'] == '.json':
                    with open(tokenizer_file_path, 'w', encoding='utf-8') as _f:
                        json.dump(saved_content['content'], _f)
                elif saved_content['file_extension'] == '.txt':
                    with open(tokenizer_file_path, 'w', encoding='utf-8') as _f:
                        for line in saved_content['content']:
                            _f.write(line)
                            _f.write('\n')
                elif saved_content['file_extension'] == '.py':
                    with open(tokenizer_file_path, 'w', encoding='utf-8') as _f:
                        _f.write(saved_content['content'])
                elif saved_content['file_extension'] == '.model':
                    try:
                        import sentencepiece as spm
                    except ImportError as e:
                        raise MissingConditionalImportError(
                            extra_deps_group='sentencepiece',
                            conda_package='sentencepiece',
                        ) from e
                    s = spm.SentencePieceProcessor()
                    s.load_from_serialized_proto(saved_content['content'])  # pyright: ignore[reportGeneralTypeIssues]
                    with open(tokenizer_file_path, 'wb') as _f:
                        _f.write(s.serialized_model_proto())

            hf_tokenizer = transformers.AutoTokenizer.from_pretrained(
                tokenizer_save_dir,
                trust_remote_code=trust_remote_code,
            )

            # we need to set the name_or_path back because otherwise it is the tmp dir we are loading from here
            # For backwards compatibility we try both the old and new key
            tokenizer_config_key = 'tokenizer_config.json' if 'tokenizer_config.json' in hf_tokenizer_state else 'tokenizer_config'
            hf_tokenizer.name_or_path = hf_tokenizer_state[tokenizer_config_key]['content'].get('name_or_path', '')
            hf_tokenizer.init_kwargs['name_or_path'] = hf_tokenizer.name_or_path

            # for an unknown reason this key is missing when loading the saved tokenizer, but present with a value of None
            # for the original tokenizer, so we default it to None
            hf_tokenizer.init_kwargs['tokenizer_file'] = hf_tokenizer.init_kwargs.get('tokenizer_file', None)
        return hf_tokenizer

    @staticmethod
    def load_huggingface_model_from_saved_state(
        hf_state: dict[str, Any],
        loaded_state_dict: dict[str, dict[str, dict[str, dict[str, Any]]]],
        model_instantiation_class: type | str | None,
        model_config_kwargs: dict[str, Any] | None,
    ) -> transformers.PreTrainedModel:
        """A helper function that loads a HuggingFace model class from a loaded in hf state.

        Args:
            hf_state (dict[str, Any]): HF state loaded from a Composer checkpoint.
            model_instantiation_class (Union[Type[:class:`transformers.PreTrainedModel`], Type[:class:`transformers.AutoModel`], str]), optional):
                Class to use to create the HuggingFace model. Defaults to the model class used in the original checkpoint. If this argument is
                a HuggingFace auto class (e.g. :class:`transformers.AutoModel` or :class:`transformers.AutoModelForSequenceClassification`), the ``from_config`` method will be used,
                while if it is of type :class:`transformers.PreTrainedModel`, the constructor will be called. This argument can also be a string,
                which will attempt to be imported as the class to use.
            model_config_kwargs: dict[str, Any]: Extra arguments to pass in for the model config creation (e.g. ``num_labels`` for creating a sequence classification model)
        Returns:
            transformers.PreTrainedModel: The loaded HuggingFace model
        """
        try:
            import transformers
        except ImportError as e:
            raise MissingConditionalImportError(
                extra_deps_group='nlp',
                conda_package='transformers',
                conda_channel='conda-forge',
            ) from e
        loaded_config = get_hf_config_from_composer_state_dict(loaded_state_dict, config_overrides=model_config_kwargs)

        hf_model_state = hf_state['model']
        if model_instantiation_class is not None:
            # If the instantiation class is explicitly provided, use it
            # If a string is provided, attempt to import the class it refers to
            if isinstance(model_instantiation_class, str):
                try:
                    model_instantiation_class = import_object(
                        ':'.join(model_instantiation_class.rsplit('.', maxsplit=1)),
                    )
                except (ModuleNotFoundError, AttributeError):
                    raise ValueError(
                        textwrap.dedent(
                            f'The provided model_instantiation_class string {model_instantiation_class} could not be imported. '
                            f'Please make sure {model_instantiation_class} is discoverable on the python path, or pass the class '
                            'in directly.',
                        ),
                    )

            assert model_instantiation_class is not None  # pyright
            # The AutoModel* classes have `from_config`, while the PreTrainedModel classes do not
            # pyright can't tell this isn't a string at this point
            if issubclass(
                model_instantiation_class,  # type: ignore
                transformers.models.auto.auto_factory._BaseAutoModelClass,  # type: ignore
            ):  # pyright: ignore[reportGeneralTypeIssues]
                hf_model = model_instantiation_class.from_config(loaded_config)  # type: ignore
            else:
                hf_model = model_instantiation_class(loaded_config)  # type: ignore
        else:
            # If the instantiation class is not explicitly provided, attempt to import the saved class and use it
            try:
                saved_class = import_object(':'.join(hf_model_state['config']['class'].rsplit('.', maxsplit=1)))
            except (ModuleNotFoundError, AttributeError):
                raise ValueError(
                    textwrap.dedent(
                        f'The saved class {hf_model_state["config"]["class"]} could not be imported. '
                        'Please either pass in the class to use explicitly via the model_instantiation_class '
                        f'parameter, or make sure that {hf_model_state["config"]["class"]} is discoverable '
                        'on the python path.',
                    ),
                )
            hf_model = saved_class(loaded_config)
        return hf_model

    @staticmethod
    def hf_from_composer_checkpoint(
        checkpoint_path: str,
        model_instantiation_class: Optional[Union[Type[transformers.PreTrainedModel],
                                                  Type['_BaseAutoModelClass'],
                                                  str,
                                                 ]] = None,
        model_config_kwargs: Optional[dict] = None,
        local_checkpoint_save_location: Optional[Union[Path, str]] = None,
        trust_remote_code: bool = False,
    ) -> tuple[transformers.PreTrainedModel,
               Optional[Union[transformers.PreTrainedTokenizer,
                              transformers.PreTrainedTokenizerFast,
                             ]],
              ]:
        """Loads a HuggingFace model (and tokenizer if present) from a composer checkpoint.

        .. note:: This function does not load the weights from the checkpoint. It just loads the correctly configured
            model and tokenizer classes.

        .. testsetup::

            import torch

            dataset = RandomTextClassificationDataset(size=16, use_keys=True)
            train_dataloader = torch.utils.data.DataLoader(dataset, batch_size=8)
            eval_dataloader = torch.utils.data.DataLoader(dataset, batch_size=8)

            import transformers
            from composer.models import HuggingFaceModel
            from composer.trainer import Trainer

            hf_model = transformers.AutoModelForSequenceClassification.from_pretrained('prajjwal1/bert-tiny', num_labels=2)
            hf_tokenizer = transformers.AutoTokenizer.from_pretrained('prajjwal1/bert-tiny')
            composer_model = HuggingFaceModel(hf_model, tokenizer=hf_tokenizer, metrics=[], use_logits=True)
            trainer = Trainer(model=composer_model,
                              train_dataloader=train_dataloader,
                              save_filename='composer-hf-checkpoint.pt',
                              max_duration='1ep',
                              save_folder='./')
            trainer.fit()
            trainer.close()

        Example:

        .. testcode::

            hf_model, hf_tokenizer = HuggingFaceModel.hf_from_composer_checkpoint('composer-hf-checkpoint.pt')
            # At this point, hf_model is randomly initialized
            composer_model = HuggingFaceModel(hf_model, hf_tokenizer)
            trainer = Trainer(model=composer_model,
                              train_dataloader=train_dataloader,
                              save_filename='composer-hf-checkpoint-2.pt',
                              max_duration='1ep',
                              save_folder='./',
                              load_path='composer-hf-checkpoint.pt')
            # At this point, the weights have been loaded from the composer checkpoint into hf_model

        Args:
            checkpoint_path (str): Path to the composer checkpoint, can be a local path, or a remote path beginning with ``s3://``, or another backend
                supported by :meth:`composer.utils.maybe_create_object_store_from_uri`.
            model_instantiation_class (Union[Type[:class:`transformers.PreTrainedModel`], Type[:class:`transformers.AutoModel`], str]), optional):
                Class to use to create the HuggingFace model. Defaults to the model class used in the original checkpoint. If this argument is
                a HuggingFace auto class (e.g. :class:`transformers.AutoModel` or :class:`transformers.AutoModelForSequenceClassification`), the ``from_config`` method will be used,
                while if it is of type :class:`transformers.PreTrainedModel`, the constructor will be called. This argument can also be a string,
                which will attempt to be imported as the class to use.
            model_config_kwargs: dict[str, Any]: Extra arguments to pass in for the model config creation (e.g. ``num_labels`` for creating a sequence classification model)
            local_checkpoint_save_location (Optional[Union[Path, str]], optional): If specified, where to save the checkpoint file to locally.
                                                                                   If the input ``checkpoint_path`` is already a local path, this will be a symlink.
                                                                                   Defaults to None, which will use a temporary file.
            trust_remote_code (bool, optional): Whether to trust the remote code when loading the tokenizer. Defaults to False.

        Raises:
            ValueError: If the ``model_instantiation_class``, or the model class saved in the checkpoint, is not able to be imported

        Returns:
            tuple[transformers.PreTrainedModel, Optional[Union[transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast]]]: The loaded HuggingFace model and (if present) tokenizer
        """

        # default local path to a tempfile if path is not provided
        if local_checkpoint_save_location is None:
            tmp_dir = tempfile.TemporaryDirectory()
            local_checkpoint_save_location = Path(tmp_dir.name) / 'local-composer-checkpoint.pt'

        if model_config_kwargs is None:
            model_config_kwargs = {}

        # download the checkpoint file
        get_file(checkpoint_path, str(local_checkpoint_save_location))

        # load the state dict in
        loaded_state_dict = safe_torch_load(local_checkpoint_save_location)

        hf_state = loaded_state_dict['state']['integrations']['huggingface']
        hf_tokenizer = HuggingFaceModel.load_huggingface_tokenizer_from_saved_state(hf_state, trust_remote_code)
        hf_model = HuggingFaceModel.load_huggingface_model_from_saved_state(
            hf_state,
            loaded_state_dict,
            model_instantiation_class,
            model_config_kwargs,
        )

        return hf_model, hf_tokenizer

    def forward(self, batch):
        if isinstance(batch, Mapping):
            # Further input validation is left to the huggingface forward call
            batch = {k: v for k, v in batch.items() if k in self.model_forward_args}
            output = self.model(**batch)  # type: ignore (thirdparty)
        else:
            raise ValueError(
                'Unexpected batch type. Expected a dictionary with keys corresponding to the inputs to the forward function of the Huggingface model',
            )
        return output

    def loss(self, outputs, batch):
        if self.config.use_return_dict:
            return outputs['loss']
        else:
            # loss is at index 0 in the output tuple
            return outputs[0]

    def eval_forward(self, batch, outputs: Optional[Any] = None):
        # If the batch mode is generate, we will generate a requested number of tokens using the underlying
        # model's generate function. Extra generation kwargs can be passed in via the batch. Strings will
        # be returned from eval_forward
        if batch.get('mode', None) == 'generate':
            if self.tokenizer is None:
                raise ValueError(
                    'Generation eval cannot be used without providing a tokenizer to the model constructor.',
                )

            self.labels = batch.pop('labels')
            generation = self.generate(
                batch['input_ids'],
                attention_mask=batch['attention_mask'],
                synced_gpus=dist.get_world_size() > 1,
                **batch.get('generation_kwargs', {}),
            )

            # don't remove prefix space to sentencepiece models
            if len(
                self.tokenizer(' a', add_special_tokens=False)['input_ids'],  # pyright: ignore[reportGeneralTypeIssues]
            ) == 1:
                return self.tokenizer.batch_decode(
                    generation[:, batch['input_ids'].shape[1]:],
                    skip_special_tokens=True,
                )
            else:
                return [
                    ' ' + generation for generation in
                    self.tokenizer.batch_decode(generation[:, batch['input_ids'].shape[1]:], skip_special_tokens=True)
                ]

        if self.use_logits or batch.get('mode', None) == 'icl_task':
            # pop labels first to avoid computing loss
            self.labels = batch.pop('labels')

            # HF encoder decoder models like T5 expect either decoder_input_ids or labels,
            # so we add decoder_input_ids to the batch if it is missing
            if self.config.is_encoder_decoder and 'decoder_input_ids' not in batch:
                if hasattr(self.model, 'prepare_decoder_input_ids_from_labels'):
                    batch['decoder_input_ids'] = self.model.prepare_decoder_input_ids_from_labels(labels=self.labels)
                else:
                    raise RuntimeError(
                        'Encoder decoder models require that either decoder_input_ids is present in the batch'
                        ' or that the model has a prepare_decoder_input_ids_from_labels method.',
                    )

            if self.shift_labels or batch.get('mode', None) == 'icl_task':
                assert self.labels is not None
                # HF CausalLM models internally shift labels before computing loss, so we do the same here
                self.labels[:, :-1] = self.labels[:, 1:].clone()
                self.labels[:, -1] = -100

            output = outputs if outputs else self.forward(batch)

            if self.config.use_return_dict:
                output = output['logits']
            else:
                # if loss was computed (cached outputs from forward), loss is at index 0 and logits are at index 1
                # if loss was not computed (no cached outputs during eval), loss is not present and logits are at index 0
                output = output[1] if len(output[0].shape) == 0 else output[0]

            # if we are in the single class case, then remove the classes dimension
            if output.ndim == 2 and output.shape[1] == 1:
                output = output.squeeze(dim=1)
        else:
            output = outputs if outputs else self.forward(batch)

        return output

    def get_metrics(self, is_train: bool = False) -> dict[str, Metric]:
        if is_train:
            metrics = self.train_metrics
        else:
            metrics = self.val_metrics

        return metrics if metrics else {}

    def update_metric(self, batch: Any, outputs: Any, metric: Metric) -> dict:
        if metric.device.type == 'cpu':
            self.labels = DeviceCPU().batch_to_device(self.labels)

        if getattr(metric, 'needs_batch', False):
            metric_result = metric.update(batch=batch, outputs=outputs, labels=self.labels)
        else:
            metric_result = metric.update(outputs, self.labels)
        if metric_result is not None:
            # Add the metric name once for each datapoint in the batch
            metric_result['metric_name'] = [metric.__class__.__name__ for _ in range(0, batch['input_ids'].shape[0])]
        else:
            metric_result = {}
        return metric_result

    def get_metadata(self):
        model_output = {}
        tokenizer_output = {}
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_dir = Path(tmp_dir)
            model_dir = tmp_dir / 'model'
            tokenizer_dir = tmp_dir / 'tokenizer'

            original_model_config: PretrainedConfig = self.config
            original_model_config.save_pretrained(model_dir)
            if self.tokenizer is not None:
                self.tokenizer.save_pretrained(tokenizer_dir)

            with open(model_dir / 'config.json') as _config_file:
                model_config = json.load(_config_file)

            model_output['config'] = {
                'file_extension': '.json',
                'content': model_config,
                'class': f'{self.model.__class__.__module__}.{self.model.__class__.__name__}',
            }

            # Also save PEFT config if the model is a peft model
            if self.using_peft:
                active_adapter = self.model.active_adapter
                assert isinstance(active_adapter, str)
                self.model.peft_config[active_adapter].save_pretrained(str(model_dir))
                with open(model_dir / 'adapter_config.json') as _peft_config_file:
                    peft_config = json.load(_peft_config_file)

                model_output['peft_config'] = {
                    'file_extension': '.json',
                    'content': peft_config,
                }

            if self.tokenizer is not None:
                for tokenizer_file_name in tokenizer_dir.iterdir():
                    tokenizer_file_path = tokenizer_dir / tokenizer_file_name
                    tokenizer_file_extension = tokenizer_file_path.suffix
                    if tokenizer_file_extension == '.txt':
                        with open(tokenizer_file_path, encoding='utf-8') as _tokenizer_file:
                            tokenizer_file_content = _tokenizer_file.read().split('\n')
                    elif tokenizer_file_extension == '.json':
                        with open(tokenizer_file_path, 'rb') as _tokenizer_file:
                            tokenizer_file_content = json.load(_tokenizer_file)
                    elif tokenizer_file_extension == '.py':
                        with open(tokenizer_file_path, encoding='utf-8') as _tokenizer_file:
                            tokenizer_file_content = _tokenizer_file.read()
                    elif tokenizer_file_extension == '.model':
                        try:
                            import sentencepiece as spm
                        except ImportError as e:
                            raise MissingConditionalImportError(
                                extra_deps_group='sentencepiece',
                                conda_package='sentencepiece',
                            ) from e
                        s = spm.SentencePieceProcessor(
                            model_file=str(tokenizer_file_path),  # pyright: ignore[reportGeneralTypeIssues]
                        )
                        tokenizer_file_content = s.serialized_model_proto()
                    else:
                        raise ValueError(
                            f'Unexpected file ending {tokenizer_file_name} in output of tokenizer.save_pretrained.',
                        )

                    tokenizer_output[tokenizer_file_path.name] = {
                        'file_extension': tokenizer_file_extension,
                        'content': tokenizer_file_content,
                    }
        return {'model': model_output, 'tokenizer': tokenizer_output}

    def generate(self, input_ids: torch.Tensor, **kwargs):
        """Generate from the underlying HuggingFace model.

        Except for ``pad_token_id``, which is optionally read from ``self.tokenizer``, all args are passed along
        to :meth:`transformers.GenerationMixin.generate` function.

        Args:
            input_ids (torch.Tensor): Input ids to generate from.
            **kwargs: Additional arguments passed to :meth:`transformers.GenerationMixin.generate` function.
                See :class:`transformers.GenerationConfig` for all available arguments.
        """
        pad_token_id = kwargs.pop('pad_token_id', self.tokenizer.pad_token_id if self.tokenizer is not None else None)

        if is_model_fsdp(self.model):
            from torch.distributed.fsdp import FullyShardedDataParallel as FSDP

            # Note: We need to use the FSDP.summon_full_params context manager here because the generate function
            # does not seem to gather the weights for the LM head. This solution works because the tied weights of the LM head
            # are in the root FSDP module, and are summoned by the below context manager. See https://github.com/pytorch/pytorch/issues/100069
            # for more info.
            # Note: We use recurse=False here so that we only summon full params for the LM head, not the entire model.
            with FSDP.summon_full_params(self.model, writeback=False, recurse=False):
                return self.model.generate(input_ids=input_ids, pad_token_id=pad_token_id, **kwargs)
        else:
            return self.model.generate(input_ids=input_ids, pad_token_id=pad_token_id, **kwargs)


def _maybe_get_peft_model(
    peft_config: 'PeftConfig',
    model: Union[transformers.PreTrainedModel, 'PeftModel'],
) -> 'PeftModel':
    """Creates a PEFT model if the model is not already a PEFT model.

    Args:
        peft_config (Optional[peft.PeftConfig]): The PEFT config to use to create the PEFT model
        model (Union[transformers.PreTrainedModel, 'PeftModel']): The model to create the PEFT model from

    Returns:
        PeftModel: The PEFT model
    """
    if not peft_installed:
        raise MissingConditionalImportError(extra_deps_group='peft', conda_package='peft', conda_channel='conda-forge')

    if not isinstance(model, PeftModel):
        log.info('Creating PEFT model')
        peft_model = get_peft_model(model, peft_config)
        assert isinstance(peft_model, PeftModel)
        return peft_model
    else:
        warnings.warn('PEFT model was passed in directly. Ignoring the provided PEFT config.')
        return model


def maybe_get_underlying_model(
    model: Union[transformers.PreTrainedModel, 'PeftModel'],
) -> Union[transformers.PreTrainedModel, 'PeftModel']:
    """Get the underlying PreTrainedModel from a model if it is a PEFT model

    Args:
        model (Union[transformers.PreTrainedModel, 'PeftModel']): The model to get the underlying model from

    Returns:
        Union[transformers.PreTrainedModel]: The underlying transformers model
    """
    if peft_installed and isinstance(model, PeftModel):
        return model.base_model.model
    else:
        return model


def _is_registered_causal_lm(model: Union[transformers.PreTrainedModel, 'PeftModel']) -> bool:
    """Return True if model class is either a registered 🤗 Causal LM or a subclass of one"""
    try:
        from transformers.models.auto.modeling_auto import MODEL_FOR_CAUSAL_LM_MAPPING
    except ImportError as e:
        raise MissingConditionalImportError(
            extra_deps_group='nlp',
            conda_package='transformers',
            conda_channel='conda-forge',
        ) from e

    model_to_check = maybe_get_underlying_model(model)

    # This try/except is needed until https://github.com/huggingface/transformers/issues/26778
    # is resolved in a release. This means that this attempt to automatically detect causal LMs
    # does not currently work in an environment with flash attention <2 installed.
    try:
        causal_lm_classes = list(MODEL_FOR_CAUSAL_LM_MAPPING.values())
    except RuntimeError as e:
        if 'Failed to import transformers.models' in str(e):
            MODEL_FOR_CAUSAL_LM_MAPPING = {}
            return False
        else:
            raise e
    return any(isinstance(model_to_check, causal_lm_class) for causal_lm_class in causal_lm_classes)  # type: ignore


def get_hf_config_from_composer_state_dict(
    state_dict: dict[str, Any],
    config_overrides: Optional[dict[str, Any]] = None,
) -> 'PretrainedConfig':
    """Get a HuggingFace config from a composer state dict with overrides applied

    Args:
        state_dict (dict[str, Any]): The state dict to get the config from
        config_overrides (dict[str, Any], optional): Any overrides to apply to the config

    Returns:
        transformers.PretrainedConfig: The HuggingFace config
    """
    try:
        import transformers
    except ImportError as e:
        raise MissingConditionalImportError(
            extra_deps_group='nlp',
            conda_package='transformers',
            conda_channel='conda-forge',
        ) from e

    if config_overrides is None:
        config_overrides = {}

    hf_config_dict = state_dict['state']['integrations']['huggingface']['model']['config']['content']
    # Update the config with any extra args needed
    hf_config_dict.update(config_overrides)
    # JSON keys need to be converted back to ints, huggingface does not auto convert them along this code path
    if 'id2label' in hf_config_dict:
        hf_config_dict['id2label'] = {int(k): v for k, v in hf_config_dict['id2label'].items()}

    try:
        return transformers.AutoConfig.for_model(**hf_config_dict)
    except ValueError:
        try:
            return transformers.AutoConfig.from_pretrained(hf_config_dict['_name_or_path'], **hf_config_dict)
        except KeyError:
            raise Exception(
                f'Could not load config from state dict using either `for_model` or `from_pretrained`.'
                f'Please make sure that the model_type={hf_config_dict.get("model_type")} is valid, or that the'
                f'config has a valid `_name_or_path`.',
            )


def get_peft_config_from_composer_state_dict(state_dict: dict[str, Any]) -> Optional['PeftConfig']:
    """Get a PEFT config from a composer state dict

    Args:
        state_dict (dict[str, Any]): The state dict to get the config from

    Returns:
        Optional[peft.PeftConfig]: The PEFT config. Will be ``None`` if the model is not a PEFT model.
    """
    try:
        import peft
    except ImportError as e:
        raise MissingConditionalImportError(
            extra_deps_group='nlp',
            conda_package='peft',
            conda_channel='conda-forge',
        ) from e

    hf_model_dict = state_dict['state']['integrations']['huggingface']['model']
    if 'peft_config' not in hf_model_dict:
        return None

    peft_config_dict = hf_model_dict['peft_config']['content']

    return peft.get_peft_config(peft_config_dict)


def write_huggingface_pretrained_from_composer_checkpoint(
    checkpoint_path: Union[Path, str],
    output_folder: Union[Path, str],
    local_checkpoint_save_location: Optional[Union[Path, str]] = None,
) -> None:
    """Write a ``config.json`` and ``pytorch_model.bin``, like :meth:`transformers.PreTrainedModel.from_pretrained` expects, from a composer checkpoint

    .. note:: This function will not work properly if you used surgery algorithms when you trained your model. In that case you will want to
        load the model weights using the Composer :class:`~composer.Trainer` with the ``load_path`` argument.

    .. testsetup::

        import torch

        dataset = RandomTextClassificationDataset(size=16, use_keys=True)
        train_dataloader = torch.utils.data.DataLoader(dataset, batch_size=8)
        eval_dataloader = torch.utils.data.DataLoader(dataset, batch_size=8)

        import transformers
        from composer.models import HuggingFaceModel
        from composer.trainer import Trainer

        hf_model = transformers.AutoModelForSequenceClassification.from_pretrained('prajjwal1/bert-tiny', num_labels=2)
        hf_tokenizer = transformers.AutoTokenizer.from_pretrained('prajjwal1/bert-tiny')
        composer_model = HuggingFaceModel(hf_model, tokenizer=hf_tokenizer, metrics=[], use_logits=True)
        trainer = Trainer(model=composer_model,
                            train_dataloader=train_dataloader,
                            save_filename='composer-hf-checkpoint.pt',
                            max_duration='1ep',
                            save_folder='./')
        trainer.fit()
        trainer.close()

    Example:

    .. testcode::

        from composer.models import write_huggingface_pretrained_from_composer_checkpoint

        write_huggingface_pretrained_from_composer_checkpoint('composer-hf-checkpoint.pt', './hf-save-pretrained-output')
        loaded_model = transformers.AutoModelForSequenceClassification.from_pretrained('./hf-save-pretrained-output')

    Args:
        checkpoint_path (Union[Path, str]): Path to the composer checkpoint, can be a local path, or a remote path beginning with ``s3://``, or another backend
            supported by :meth:`composer.utils.maybe_create_object_store_from_uri`.
        output_folder (Union[Path, str]): Path to the folder to write the output to. Must be a local path.
        local_checkpoint_save_location (Optional[Union[Path, str]], optional): If specified, where to save the checkpoint file to locally.
                                                                                If the input ``checkpoint_path`` is already a local path, this will be a symlink.
                                                                                Defaults to None, which will use a temporary file.
    """
    try:
        import transformers
        del transformers
    except ImportError as e:
        raise MissingConditionalImportError(
            extra_deps_group='nlp',
            conda_package='transformers',
            conda_channel='conda-forge',
        ) from e

    # default local path to a tempfile if path is not provided
    if local_checkpoint_save_location is None:
        tmp_dir = tempfile.TemporaryDirectory()
        local_checkpoint_save_location = Path(tmp_dir.name) / 'local-composer-checkpoint.pt'

    # download the checkpoint file
    get_file(str(checkpoint_path), str(local_checkpoint_save_location))

    composer_state_dict = safe_torch_load(local_checkpoint_save_location)

    # load tokenizer
    hf_state = composer_state_dict['state']['integrations']['huggingface']
    hf_tokenizer = HuggingFaceModel.load_huggingface_tokenizer_from_saved_state(hf_state)
    assert hf_tokenizer is not None
    hf_tokenizer.save_pretrained(output_folder)

    config = get_hf_config_from_composer_state_dict(composer_state_dict)
    config.save_pretrained(output_folder)

    peft_config = get_peft_config_from_composer_state_dict(composer_state_dict)
    if peft_config is not None:
        peft_config.save_pretrained(str(output_folder))

    weights_state_dict = composer_state_dict['state']['model']
    torch.nn.modules.utils.consume_prefix_in_state_dict_if_present(weights_state_dict, prefix='model.')

    # NOTE: This only works for default adapter name, not multiple adapters
    if peft_config is not None:
        weights_state_dict = filter_state_dict_peft(weights_state_dict, peft_config, adapter_name='default')

        torch.save(weights_state_dict, Path(output_folder) / 'adapter_model.bin')
    else:
        torch.save(weights_state_dict, Path(output_folder) / 'pytorch_model.bin')


def filter_state_dict_peft(
    state_dict: dict[str, Any],
    peft_config: 'PeftConfig',
    adapter_name: str = 'default',
    remove_adapter_names: bool = True,
) -> dict[str, Any]:
    """Filter a state dict to only include the weights needed for a PEFT model

    Note: This function only works with LORA PEFT models right now.

    Args:
        state_dict (dict[str, Any]): The state dict to filter
        peft_config (PeftConfig): The PEFT config to use to filter the state dict
        adapter_name (str, optional): The name of the adapter to filter for. Defaults to 'default'.
        remove_adapter_names (bool, optional): Whether to remove the adapter names from the state dict keys. Defaults to True.

    Returns:
        dict[str, Any]: The filtered state dict
    """

    if peft_config.peft_type != 'LORA':
        raise NotImplementedError(f'Only LoRA PEFT is supported. Got {peft_config.peft_type}')

    # Filtering copied from https://github.com/huggingface/peft/blob/4186c9b104644fd247a4cc0dc2dfc1ede4665204/src/peft/utils/save_and_load.py#L68C1-L86C116
    bias = peft_config.bias  # type: ignore
    if bias == 'none':
        to_return = {k: state_dict[k] for k in state_dict if 'lora_' in k}
    elif bias == 'all':
        to_return = {k: state_dict[k] for k in state_dict if 'lora_' in k or 'bias' in k}
    elif bias == 'lora_only':
        to_return = {}
        for k in state_dict:
            if 'lora_' in k:
                to_return[k] = state_dict[k]
                bias_name = k.split('lora_')[0] + 'bias'
                if bias_name in state_dict:
                    to_return[bias_name] = state_dict[bias_name]
    else:
        raise NotImplementedError
    to_return = {k: v for k, v in to_return.items() if (('lora_' in k and adapter_name in k) or ('bias' in k))}

    if remove_adapter_names:
        to_return = {k.replace(f'.{adapter_name}', ''): v for k, v in to_return.items()}
    return to_return
